import re

from bs4 import BeautifulSoup

from h2m2t.html import fix_html_tree
from h2m2t.markdown import MarkdownConverterFixed


def html_to_markdown(
    html: str,
    parser: str = "html.parser",
    *,
    strip_tags: list | None = None,
) -> str:
    soup = BeautifulSoup(html, parser)
    fix_html_tree(soup)

    converter = MarkdownConverterFixed(strip=strip_tags, autolinks=False, heading_style="atx")
    md = converter.convert_soup(soup)
    md = f"\n{md}\n"

    # Fix spaces after inline elements splitting
    md = re.sub("\u00a0{2,}", "\u00a0", md, flags=re.UNICODE)
    md = md.replace(" \u00a0", " ").replace("\u00a0 ", " ")
    md = md.replace("\u00a0\n", "\n").replace("\n\u00a0", "\n")

    # Clean newlines
    md = re.sub(r"\n[^\S\n]+\n", "\n\n", md, flags=re.UNICODE)
    md = re.sub(r"\n[^\S\n]+\n", "\n\n", md, flags=re.UNICODE)
    md = re.sub(r"\n{3,}", "\n\n", md)

    return md.strip()



import re
from html import unescape

from ftfy import fix_text

whitespace_chars = [
    # "\u0009",
    # "\u000a",
    # "\u000b",
    # "\u000c",
    # "\u000d",
    # "\u0020",
    # "\u0085",
    # "\u00a0",
    # "\u1680",
    # "\u2000",
    # "\u2001",
    # "\u2002",
    # "\u2003",
    # "\u2004",
    # "\u2005",
    # "\u2006",
    # "\u2007",
    # "\u2008",
    # "\u2009",
    # "\u200a",
    "\u200b",
    "\u200c",
    "\u200d",
    # "\u2028",
    # "\u2029",
    # "\u202f",
    # "\u205f",
    "\u2060",
    # "\u3000",
]


def repair_html_entities(text: str) -> str:
    before = ""

    while text != before:
        before, text = text, unescape(text)

    return text


def repair_broken_unicode(text: str) -> str:
    return fix_text(text, normalization="NFKD")


def replace_rare_spaces(text: str) -> str:
    spaces = "".join(whitespace_chars)

    return re.sub(f"[{spaces}]", " ", text, flags=re.UNICODE)



from h2m2t.h2m import html_to_markdown
from h2m2t.m2t import markdown_to_text


def html_to_text(
    html: str,
    parser: str = "html.parser",
    *,
    strip_tags: list | None = None,
) -> str:
    md = html_to_markdown(html, parser, strip_tags=strip_tags)
    txt = markdown_to_text(md)

    return txt.strip()



import re
from collections.abc import Callable

from bs4 import Tag
from markdownify import MarkdownConverter
from mistune import HTMLRenderer

from h2m2t.text import replace_rare_spaces


class HTMLRendererFixed(HTMLRenderer):
    def link(self, text: str, url: str, title: str | None = None) -> str:
        url = url.replace("%20", " ")
        url = url.replace("%28", "(").replace("%29", ")")
        return super().link(text, url, title)


class MarkdownConverterFixed(MarkdownConverter):
    root_tags = (
        "a",
        "b",
        "blockquote",
        "code",
        "pre",
        "del",
        "em",
        "kbd",
        "dd",
        "dl",
        "dt",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "i",
        "ul",
        "ol",
        "table",
        "s",
        "strong",
        "samp",
        "sub",
        "sup",
    )

    def process_text(self, el: Tag, parent_tags: set[str] | None = None) -> str:
        text = super().process_text(el, parent_tags)

        if "code" not in parent_tags and "pre" not in parent_tags:
            text = text.replace("<", "\\<").replace(">", "\\>")

        if set(self.root_tags).intersection(parent_tags):
            return text

        for b in self.options["bullets"]:
            if text.rstrip() == b or text.startswith(f"{b} "):
                text = f"\\{text}"

        if m := re.match(r"\d+\. ", text):
            number, text = text[: m.end() - 2], text[m.end() :]
            text = rf"{number}\. {text}"

        return text

    def get_conv_fn(self, tag_name: str) -> Callable[[Tag, str, set[str]], str]:
        conv_fn = super().get_conv_fn(tag_name)

        def _conv_fn(el: Tag, text: str, parent_tags: set[str]) -> str:
            text = replace_rare_spaces(text)
            return conv_fn(el, text, parent_tags)

        return conv_fn if conv_fn is None else _conv_fn

    def _convert_hn(self, n: int, el: Tag, text: str, parent_tags: set[str]) -> str:
        if not text.strip():
            return "\n"

        text = re.sub("\n+", "\t", text).strip()

        if "_inline" not in parent_tags and set(self.root_tags).intersection(parent_tags):
            return "\n\n" + self.convert_b(el, text, parent_tags).strip() + "\n\n"

        return super()._convert_hn(n, el, text, parent_tags)

    def convert_pre(self, el: Tag, text: str, parent_tags: set[str]) -> str:
        text = text.strip("\n")

        return super().convert_pre(el, text, parent_tags)

    def convert_code(self, el: Tag, text: str, parent_tags: set[str]) -> str:
        text = text.strip("\n")

        return super().convert_code(el, text, parent_tags)

    def convert_img(self, el: Tag, text: str, parent_tags: set[str]) -> str:
        self.strip_attr(el, "alt")
        self.strip_attr(el, "title")
        self.strip_attr(el, "src")
        self.strip_attr(el, "data-original")

        alt = el.attrs["alt"]
        title = el.attrs["title"]
        src = el.attrs["src"]
        original = el.attrs["data-original"]

        if not src and original:
            el.attrs["src"] = src = original

        if not alt and not title and not src:
            return ""

        return super().convert_img(el, text, parent_tags)

    def convert_video(self, el: Tag, text: str, parent_tags: set[str]) -> str:
        self.strip_attr(el, "src")
        self.strip_attr(el, "poster")

        return super().convert_video(el, text, parent_tags)

    def convert_a(self, el: Tag, text: str, parent_tags: set[str]) -> str:
        self.strip_attr(el, "title")
        self.strip_attr(el, "href")

        href = el.attrs["href"].replace(" ", "%20")
        href = href.replace("(", "%28").replace(")", "%29")
        el.attrs["href"] = href

        text = re.sub("\n+", "\t", text).strip()

        return super().convert_a(el, text, parent_tags)

    def convert_option(self, el: Tag, text: str, parent_tags: set[str]) -> str:
        return super().convert_div(el, text, parent_tags)

    @staticmethod
    def strip_attr(el: Tag, attr: str) -> None:
        value = el.attrs.get(attr, None) or ""
        value = replace_rare_spaces(value)
        value = re.sub("\n+", "\t", value)
        el.attrs[attr] = value.strip()



from mistune import create_markdown

from h2m2t.markdown import HTMLRendererFixed


def markdown_to_html(md: str) -> str:
    renderer = HTMLRendererFixed(escape=False, allow_harmful_protocols=True)
    markdown = create_markdown(
        renderer=renderer,
        plugins=[
            # Default:
            "strikethrough",
            "footnotes",
            "table",
            "speedup",
            # Enabled:
            "url",
            "task_lists",
            "def_list",
            "abbr",
            "mark",
            "insert",
            "superscript",
            "subscript",
            "math",
            # "ruby",
            "spoiler",
        ],
    )

    markdown.block.rules.remove("setex_heading")
    markdown.block.block_quote_rules.remove("setex_heading")
    markdown.block.list_rules.remove("setex_heading")

    html = markdown(md)

    return f'<!doctype html><html><head><meta charset="utf-8"><title></title></head><body>{html}</body></html>'



from bs4 import BeautifulSoup, Comment, NavigableString, PageElement, Tag

from h2m2t.table import tables_to_dense
from h2m2t.text import replace_rare_spaces

meaningless_tags = [
    "air",  # vc.ru
    "air-settings",  # vc.ru
    "applet",
    "audio",
    "canvas",
    "embed",
    "frame",
    "frameset",
    "head",
    "iframe",
    "link",
    "map",
    "meta",
    "noembed",
    "noframes",
    "noscript",
    "object",
    "plaintext",
    "script",
    "style",
    "svg",
    "template",
    "title",
    "video",
    "xmp",
]

block_tags = [
    "address",
    "article",
    "aside",
    "blockquote",
    "br",  # inline, but behaves as block
    "canvas",
    # "dd",  # child
    "div",
    "dl",
    # "dt",  # child
    "fieldset",
    "figcaption",
    "figure",
    "footer",
    "form",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "header",
    "hr",
    "legend",
    "li",  # child
    "main",
    "nav",
    "noscript",
    "ol",
    "optgroup",  # child
    "option",  # child
    "pre",
    "section",
    "table",
    "tbody",  # child
    "td",  # child
    "tfoot",  # child
    "th",  # child
    "thead",  # child
    "tr",  # child
    "ul",
    "video",
]

inline_tags = [
    "a",
    "abbr",
    "acronym",
    "b",
    "bdo",
    "big",
    "button",
    # "br",  # inline, but behaves as block
    "cite",
    "code",
    "dfn",
    "em",
    "i",
    "img",
    "input",
    "kbd",
    "label",
    "map",
    "object",
    "output",
    "q",
    "s",
    "samp",
    "script",
    "select",
    "small",
    "span",
    "strong",
    "sub",
    "sup",
    "textarea",
    "time",
    "tt",
    "u",
    "var",
]

child_tags = [
    "dd",  # child
    "dt",  # child
    "li",  # child
    "optgroup",  # child
    "option",  # child
    "tbody",  # child
    "td",  # child
    "tfoot",  # child
    "th",  # child
    "thead",  # child
    "tr",  # child
]


def fix_html_tree(soup: BeautifulSoup) -> None:
    _drop_meaningless(soup)
    _unwrap_nested_tables(soup)
    _fix_spaces(soup)
    _fix_newlines(soup)
    _fix_inline(soup)
    _fix_block(soup)
    tables_to_dense(soup, deduplicate=True, unwrap=True)


def _drop_meaningless(soup: BeautifulSoup) -> None:
    # Drop non-meaning tags
    for node in soup(meaningless_tags):
        _merge_siblings(node)
        node.decompose()

    # Drop comments
    for comment in soup(string=lambda s: isinstance(s, Comment)):
        _merge_siblings(comment)
        comment.decompose()


def _unwrap_nested_tables(soup: BeautifulSoup) -> None:
    for table in soup("table"):
        if not table("table"):
            continue

        nodes = [table]
        for body in table(["thead", "tbody", "tfoot"], recursive=False):
            nodes.append(body)
            for row in body("tr", recursive=False):
                nodes.append(row)
                nodes.extend(row(["td", "th"], recursive=False))
        for row in table("tr", recursive=False):
            nodes.append(row)
            nodes.extend(row(["td", "th"], recursive=False))

        for node in nodes:
            node.name = "div"


def _fix_spaces(soup: BeautifulSoup) -> None:
    # Move spaces out from inline elements
    for node in soup(inline_tags):
        _fix_inline_spaces(node)


def _fix_inline_spaces(node: Tag) -> None:
    if node.contents:
        _fix_spaces_left(node, node.contents[0])
    if node.contents:
        _fix_spaces_right(node, node.contents[-1])

    if node.parent and node.parent.name in inline_tags:
        _fix_inline_spaces(node.parent)


def _fix_spaces_left(node: Tag, element: PageElement) -> None:
    if not isinstance(element, NavigableString):
        return

    body = str(element).lstrip()
    if str(element) == body:
        return

    if not body:
        node.insert_after(NavigableString(str(element)))
        element.decompose()
        return

    head = str(element).split(body)[0]
    before = node.previous_sibling
    if isinstance(before, NavigableString):
        string = str(before) + head
        before.replace_with(NavigableString(string))
    else:
        node.insert_before(NavigableString(head))

    element.replace_with(NavigableString(body))


def _fix_spaces_right(node: Tag, element: PageElement) -> None:
    if not isinstance(element, NavigableString):
        return

    body = str(element).rstrip()
    if str(element) == body:
        return

    if not body:
        node.insert_after(NavigableString(str(element)))
        element.decompose()
        return

    tail = str(element).split(body)[-1]
    after = node.next_sibling
    if isinstance(after, NavigableString):
        string = tail + str(after)
        after.replace_with(NavigableString(string))
    else:
        node.insert_after(NavigableString(tail))

    element.replace_with(NavigableString(body))


def _fix_newlines(soup: BeautifulSoup) -> None:
    # Remove linebreaks inside text nodes (as browser does)
    for node in soup(string=lambda s: isinstance(s, NavigableString) and "\n" in s):
        parents = {parent.name for parent in node.parents}
        if "code" in parents or "pre" in parents:
            continue

        string = str(node).replace("\n", " ")
        string = replace_rare_spaces(string)
        node.replace_with(NavigableString(string))


def _fix_inline(soup: BeautifulSoup) -> None:
    # Split inline elements
    for node in soup(inline_tags):
        before = node.previous_sibling
        if isinstance(before, NavigableString) and str(before) == str(before).rstrip():
            before.replace_with(NavigableString(f"{before}\u00a0"))
        if isinstance(before, Tag) and before.name in inline_tags:
            node.insert_before(NavigableString("\u00a0"))

        after = node.next_sibling
        if isinstance(after, NavigableString) and str(after) == str(after).lstrip():
            after.replace_with(NavigableString(f"\u00a0{after}"))
        if isinstance(after, Tag) and after.name in inline_tags:
            node.insert_after(NavigableString("\u00a0"))


def _fix_block(soup: BeautifulSoup) -> None:
    # Remove spaces at block element borders
    for node in soup(block_tags):
        if node.name in {"code", "pre"}:
            continue

        _fix_block_left(node)
        _fix_block_right(node)

        # Split block element
        if node.name != "br" and node.name not in child_tags:
            if node.previous_sibling:
                node.insert_before(soup.new_tag("br"))
            if node.next_sibling:
                node.insert_after(soup.new_tag("br"))


def _fix_block_left(node: Tag) -> None:
    # Around block element
    if isinstance(node.previous_sibling, NavigableString):
        string = str(node.previous_sibling).rstrip()
        if string:
            node.previous_sibling.replace_with(NavigableString(string))
        else:
            node.previous_sibling.decompose()

    # Inside block element
    if node.contents and isinstance(node.contents[0], NavigableString):
        string = str(node.contents[0]).lstrip()
        if string:
            node.contents[0].replace_with(NavigableString(string))
        else:
            node.contents[0].decompose()


def _fix_block_right(node: Tag) -> None:
    # Around block element
    if isinstance(node.next_sibling, NavigableString):
        string = str(node.next_sibling).lstrip()
        if string:
            node.next_sibling.replace_with(NavigableString(string))
        else:
            node.next_sibling.decompose()

    # Inside block element
    if node.contents and isinstance(node.contents[-1], NavigableString):
        string = str(node.contents[-1]).rstrip()
        if string:
            node.contents[-1].replace_with(NavigableString(string))
        else:
            node.contents[-1].decompose()


def _merge_siblings(element: PageElement) -> None:
    before = element.previous_sibling
    if isinstance(before, Comment):
        return
    if not isinstance(before, NavigableString):
        return

    after = element.next_sibling
    if isinstance(after, Comment):
        return
    if not isinstance(after, NavigableString):
        return

    string = str(before) + str(after)
    before.replace_with(NavigableString(string))
    after.decompose()



import re
from typing import Any, ClassVar, Literal

from bs4 import BeautifulSoup
from mistune import BaseRenderer, BlockState, InlineParser, Markdown, import_plugin


def markdown_to_text(md: str, parser: str = "html.parser") -> str:
    plugins = [
        # Default:
        "strikethrough",
        "footnotes",
        "table",
        "speedup",
        # Enabled:
        "url",
        "task_lists",
        "def_list",
        "abbr",
        "mark",
        "insert",
        "superscript",
        "subscript",
        "math",
        # "ruby",
        "spoiler",
    ]

    inline = InlineParser(hard_wrap=False)
    real_plugins = [import_plugin(n) for n in plugins]
    markdown = Markdown(renderer=TextRenderer(parser), inline=inline, plugins=real_plugins)

    text = markdown(md)

    text = re.sub(r"[^\S\n]*\n[^\S\n]*", "\n", text, flags=re.UNICODE)
    text = re.sub("([^\\S\n])[^\\S\n]+", " ", text, flags=re.UNICODE)
    text = re.sub("\n{3,}", "\n\n", text)

    return text.strip()


class TextRenderer(BaseRenderer):
    """A renderer for converting Markdown to Text."""

    NAME: ClassVar[Literal["text"]] = "text"

    def __init__(self, parser: str = "html.parser") -> None:
        super().__init__()
        self._html_parser = parser

    def render_token(self, token: dict[str, Any], state: BlockState) -> str:
        # backward compitable with v2
        func = self._get_method(token["type"])
        attrs = token.get("attrs")

        if "raw" in token:
            text = token["raw"]
        elif "children" in token:
            text = self.render_tokens(token["children"], state)
        elif attrs:
            return func(**attrs)
        else:
            return func()
        if attrs:
            return func(text, **attrs)
        return func(text)

    def text(self, text: str) -> str:
        return text

    def emphasis(self, text: str) -> str:
        return text

    def strong(self, text: str) -> str:
        return text

    def link(self, text: str, url: str, title: str | None = None) -> str:
        if title and title != text:
            return f"{text} [{title}]"

        return text

    def image(self, text: str, url: str, title: str | None = None) -> str:
        if title and title != text:
            return f"{text} [{title}]"

        return text

    def codespan(self, text: str) -> str:
        return text

    def linebreak(self) -> str:
        return "\n"

    def softbreak(self) -> str:
        return "\n"

    def inline_html(self, html: str) -> str:
        return BeautifulSoup(html, self._html_parser).get_text(separator=" ")

    def paragraph(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def heading(self, text: str, level: int, **attrs: Any) -> str:
        return f"\n\n{text}\n\n"

    def blank_line(self) -> str:
        return "\n"

    def thematic_break(self) -> str:
        return "\n"

    def block_text(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def block_code(self, code: str, info: str | None = None) -> str:
        return f"\n\n{code}\n\n"

    def block_quote(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def block_html(self, html: str) -> str:
        text = BeautifulSoup(html, self._html_parser).get_text(separator=" ")
        return f"\n\n{text.strip()}\n\n"

    def block_error(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def list(self, text: str, ordered: bool, **attrs: Any) -> str:
        return f"\n\n{text}\n\n"

    def list_item(self, text: str) -> str:
        return f"{text.strip()}\n"

    # strikethrough
    def strikethrough(self, text: str) -> str:
        return text

    # footnotes
    def footnote_ref(self, key: str, index: int) -> str:
        return ""

    def footnote_item(self, text: str, index: int) -> str:
        return f"{text.strip()}\n"

    def footnotes(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    # table
    def table(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def table_head(self, text: str) -> str:
        return f"\n{text}\n"

    def table_body(self, text: str) -> str:
        return f"\n{text}\n"

    def table_row(self, text: str) -> str:
        return f"{text}\n"

    def table_cell(self, text: str, align: str | None = None, head: bool = False) -> str:
        return f"\t{text.strip()}\t"

    # - speedup

    # - url

    # task_lists
    def task_list_item(self, text: str) -> str:
        return f"{text.strip()}\n"

    # def_list
    def def_list(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def def_list_head(self, text: str) -> str:
        return f"{text.strip()}\n"

    def def_list_item(self, text: str) -> str:
        return f"{text.strip()}\n"

    # abbr
    def abbr(self, text: str, title: str) -> str:
        if title and title != text:
            return f"{text} [{title}]"

        return text

    # mark
    def mark(self, text: str) -> str:
        return text

    # insert
    def insert(self, text: str) -> str:
        return text

    # superscript
    def superscript(self, text: str) -> str:
        return text

    # subscript
    def subscript(self, text: str) -> str:
        return text

    # math
    def block_math(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def inline_math(self, text: str) -> str:
        return text

    # spoiler
    def block_spoiler(self, text: str) -> str:
        return f"\n\n{text}\n\n"

    def inline_spoiler(self, text: str) -> str:
        return text



import copy
from itertools import groupby, pairwise

from bs4 import BeautifulSoup, Tag

# According to https://html.spec.whatwg.org/multipage/tables.html#attributes-common-to-td-and-th-elements
# and https://developer.mozilla.org/en-US/docs/Web/HTML/Reference/Elements/td#attributes
max_rowspan: int = 65534
max_colspan: int = 1000


def tables_to_dense(soup: BeautifulSoup, *, deduplicate: bool, unwrap: bool) -> None:
    for table in reversed(soup("table")):
        _normalize_table_parts(table, deduplicate=deduplicate)
        if unwrap:
            _unwrap_empty_table(table)


def _normalize_table_parts(table: Tag, *, deduplicate: bool) -> None:
    cells: list[list[Tag]] = []

    # Fix cell spans for each part separately
    for thead in table("thead", recursive=False):
        cells.extend(_fix_cell_spans(thead("tr", recursive=False)))

    body: list[Tag] = []
    for tbody_or_tr in table(["tbody", "tr"], recursive=False):
        if tbody_or_tr.name == "tbody":
            cells.extend(_fix_cell_spans(body))
            body = []
            cells.extend(_fix_cell_spans(tbody_or_tr("tr", recursive=False)))
        else:
            body.append(tbody_or_tr)
    cells.extend(_fix_cell_spans(body))

    for tfoot in table("tfoot", recursive=False):
        cells.extend(_fix_cell_spans(tfoot("tr", recursive=False)))

    # Build virtual grid with references to cells
    grid = _build_virtual_grid(cells)

    if deduplicate:
        # Drop repeating rows
        grid = _cleanup_virtual_grid(grid, transpose=False)
        # ... and columns
        grid = _cleanup_virtual_grid(grid, transpose=True)

    # Group rows by original parent group name
    parts = {part: list(rows) for part, rows in groupby(grid, lambda r: _estimate_rows_parent(r, cells))}

    # Move caption before table
    for caption in table("caption", recursive=False):
        caption.name = "div"
        table.insert_before(caption.extract())

    # Build new table content
    tparts = []
    for part in ["thead", "tbody", "tfoot"]:
        tbody = Tag(name=part)
        for row in parts.get(part, []):
            tr = Tag(name="tr")
            for rc in row:
                td = copy.deepcopy(cells[rc[0]][rc[1]]) if rc else Tag(name="td")
                tr.append(td)
            if tr.contents:
                tbody.append(tr)
        if tbody.contents:
            tparts.append(tbody)

    # Replace table content
    table.clear()
    table.extend(tparts)


def _fix_cell_spans(rows: list[Tag]) -> list[list[Tag]]:
    grid: list[list[Tag]] = [r(["th", "td"], recursive=False) for r in rows]

    for r, row in enumerate(grid):
        for cell in row:
            rowspan = int(cell.get("rowspan", "1") or "1")
            rowspan = 1 if rowspan < 0 else rowspan
            rowspan = max_rowspan if rowspan == 0 else rowspan
            rowspan = min(rowspan, len(rows) - r)
            cell["rowspan"] = str(rowspan)

            colspan = int(cell.get("colspan", "1") or "1")
            colspan = 1 if colspan < 1 or colspan > max_colspan else colspan
            colspan = min(max(colspan, 1), max_colspan)
            cell["colspan"] = str(colspan)

    return grid


def _build_virtual_grid(cells: list[list[Tag]]) -> list[list[tuple[int, int] | None]]:
    grid: list[list[tuple[int, int] | None]] = [[] for _ in range(len(cells))]

    for r, row in enumerate(cells):
        for c, cell in enumerate(row):
            rowspan = int(cell.get("rowspan"))
            colspan = int(cell.get("colspan"))
            del cell["rowspan"], cell["colspan"]

            for ri in range(r, len(cells)):
                grid[ri].extend([None] * colspan)

            c_empty = grid[r].index(None)

            for rs in range(r, r + rowspan):
                for cs in range(c_empty, c_empty + colspan):
                    grid[rs][cs] = (r, c)

    maxlen = max([max([c for c, cell in enumerate(row) if cell], default=-1) for row in grid], default=-1) + 1
    return [row[:maxlen] + [None] * (maxlen - len(row)) for row in grid]


def _cleanup_virtual_grid(
    grid: list[list[tuple[int, int] | None]],
    *,
    transpose: bool,
) -> list[list[tuple[int, int] | None]]:
    if transpose:
        grid = [list(row) for row in zip(*grid, strict=True)]

    equals = [i for i, (row0, row1) in enumerate(pairwise(grid)) if row0 == row1]
    for i in reversed(equals):
        del grid[i + 1]

    if transpose:
        grid = [list(row) for row in zip(*grid, strict=True)]

    return grid


def _unwrap_empty_table(table: Tag) -> None:
    def _drop_table() -> None:
        caption = table.find("caption", recursive=False)
        if caption:
            caption = copy.deepcopy(caption)
            caption.name = "div"
            table.insert_before(caption)
        table.decompose()

    rows = table("tr", recursive=False)
    for body in table(["thead", "tbody", "tfoot"], recursive=False):
        rows.extend(body("tr", recursive=False))
    cells = [cell for row in rows for cell in row(["th", "td"], recursive=False)]
    cells = [cell for cell in cells if cell.text.strip()]

    if not cells:
        _drop_table()
    elif len(cells) == 1:
        cell = copy.deepcopy(cells[0])
        cell.name = "div"
        table.insert_after(cell)
        _drop_table()


def _estimate_rows_parent(row: list[tuple[int, int] | None], cells: list[list[Tag]]) -> str:
    for rc in row:
        if not rc:
            continue
        cell = cells[rc[0]][rc[1]]
        parent = cell.find_parent(["thead", "tbody", "tfoot", "table"])

        if not parent:
            continue

        if parent.name in {"thead", "tbody", "tfoot"}:
            return parent.name

    return "tbody"



from h2m2t.h2m import html_to_markdown
from h2m2t.h2t import html_to_text
from h2m2t.m2h import markdown_to_html
from h2m2t.m2t import markdown_to_text

__all__ = [
    "html_to_markdown",
    "html_to_text",
    "markdown_to_html",
    "markdown_to_text",
]
